// Analyze source, dynamic

#include "globals.hh"

#include <iostream>
#include <ctype.h>
#include <cmath>

// no limits!
unsigned int max_word_order = 0xffffffff;

hash_map<deque<SymbolID>, unsigned int, deque_hash> counts;
hash_map<deque<SymbolID>, unsigned int, deque_hash> totalcounts;

class CharReader
{
public:
  static const char *name()
  {
    return "char reader";
  }

  static bool readSymbol(istream &is, Symbol &s)
  {
    while(is.good())
      {
	char c = is.get();
	
	// return useful characters directly
	if (isalpha(c))
	  {
	    s = char2symbol(tolower(c));
	    return true;
	  }
	// if whitespace
	else if (isspace(c))
	  {
	    // ignore more whitespaces or garbage
	    while (is.good() && !isalpha(is.peek())) is.get();
	    
	    s = char2symbol('_');
	    return true;
	  }
	
	// everything else: ignore.
      }
    
    // eof
    return false;
  }  
};

class WordReader
{
  static Symbol strip(Symbol const &s)
  {
    Symbol retval;
    
    // clean s from all unknown characters.
    // converts to lower case
    for (Symbol::const_iterator i = s.begin(); i != s.end(); i++)
      {
	if (isalpha(*i))
	  retval.push_back(tolower(*i));
      }
    
    return retval;
  }

public:
  static const char *name()
  {
    return "word reader";
  }

  static bool readSymbol(istream &is, Symbol &s)
  {
    while (true)
      {
	// ignore whitespaces
	while (is.good() && isspace(is.peek())) is.get();
	
	if (is.good())
	  is >> s;
	else
	  return false;
	
	// clean s from all strange characters (,.- etc)
	s = strip(s);
	
	// if something left of s, return it.
	if (s != Symbol(""))
	  return true;
      }	    
  }
};

template<class reader>
int analyze(unsigned int n, istream &is, ostream &os)
{
  if (n > max_word_order)
    {
      cerr << "Cannot analyse word source for more than order " 
	   << max_word_order << endl;
      return error_code_max_order;
    }

  cout << "Analyzing source using " << reader::name() 
       << ", dynamic model order " << n << endl;

  // create memory
  deque<SymbolID> memory(n+1);

  cout << "Reading stream..." << endl;

  Symbol s;
  unsigned int wordcount = 0;

  // initialize memory (get the first n Symbols)
  for (unsigned int i = 0; i < n; i++)
    {
      // already at the end?
      if (!reader::readSymbol(is, s))
	{
	  cerr << "Input stream too short, nothing done" << endl;
	  return error_code_stream_too_short;
	}

      memory.pop_front();
      memory.push_back(getSymbolID(s));
      
      wordcount++;
    }

  do
    {
      // stream finished?
      if (!reader::readSymbol(is, s))
	break;

      // remove oldest symbol from memory
      memory.pop_front();

      // add to totalcounts for this prerequisite
      totalcounts[memory]++;

      // store the new symbol in memory
      memory.push_back(getSymbolID(s));

      // add to counts for this combination
      counts[memory]++;

      wordcount++;
    }
  while (true);

  // evaluate + output
  cout << "Probabilities computed (text length: " << wordcount 
       << ", " << id2symbol.size() << " symbols)." 
       << endl << "Writing to file..." << endl;
  // first output type of analysis done
  os << "dynamic " << n << endl;
  // output 
  // number of ids
  os << id2symbol.size() << endl;
  // output word/id table
  for (unsigned int i = 0; i < id2symbol.size(); i++)
    os << i << '\t' << id2symbol[i] << endl;

  // entropy will be computed if we had 0th order only!
  float entropy = 0;

  for (hash_map<deque<SymbolID>, unsigned int, deque_hash>::iterator 
	 i = counts.begin(); i != counts.end(); i++)
    {
      // remove last element to address totalcounts
      deque<SymbolID> copy = i->first;
      copy.pop_back();

      float probability = float(i->second)/float(totalcounts[copy]);

      if (n == 0)
	entropy += probability * -log2(probability);

      os << i->first << "\t\t" << probability << endl;
    }

  // if this was 0th order, print entropy
  if (n == 0)
    {
      cout << "Entropy: " << entropy << endl;
    }

  return 0;
}

// explicit instantiation
template int analyze<CharReader>(unsigned int n, istream &is, ostream &os);
template int analyze<WordReader>(unsigned int n, istream &is, ostream &os);
